"""Automatic pyfai integration for every scan with saving and plotting"""

from typing import List, Optional
import os
from ..resources.xrpd import RESOURCE_ROOT

from ..xrpd.processor import XrpdProcessor
from ..persistent import ParameterInfo


class Id22XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
    ],
):
    def __init__(self, **defaults) -> None:
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_ocl_gpu",
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )
        defaults.setdefault(
            "workflow_with_saving",
            os.path.join(RESOURCE_ROOT, "Sum_then_integrate_with_saving.json"),
        )
        defaults.setdefault("workflow_without_saving", None)
        super().__init__(**defaults)

        # Disable data from memory for now
        # The data structure is indeed different when getting it from file or from memory
        self._set_parameter("data_from_memory", False)

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.pyfai_config

    def get_integration_options(self, lima_name: str) -> Optional[dict]:
        integration_options = self.integration_options
        if integration_options:
            return integration_options.to_dict()
        return None

    def get_inputs(self, scan, lima_name: str) -> List[dict]:
        inputs = super().get_inputs(scan, lima_name)
        inputs += self.get_sum_inputs(
            scan,
            lima_name,
        )
        inputs += self.get_save_ascii_inputs(scan, lima_name)
        return inputs

    def get_sum_inputs(self, scan, lima_name: str):
        task_identifier = "SumBlissScanImages"

        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        inputs = [
            {
                "task_identifier": task_identifier,
                "name": "filename",
                "value": filename,
            },
            {
                "task_identifier": task_identifier,
                "name": "output_filename",
                "value": self.save_nexus_filename(scan),
            },
            {
                "task_identifier": task_identifier,
                "name": "scan",
                "value": scan_nb,
            },
            {
                "task_identifier": task_identifier,
                "name": "detector_name",
                "value": lima_name,
            },
            {
                "task_identifier": task_identifier,
                "name": "skip_dark",
                "value": True,
            },
        ]
        if self.data_from_memory:
            scan_memory_url = f"{scan.root_node.db_name}:{scan._node_name}"
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "scan_memory_url",
                    "value": scan_memory_url,
                }
            )
        if self.retry_timeout:
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "retry_timeout",
                    "value": self.retry_timeout,
                }
            )
        if self.flush_period:
            inputs.append(
                {
                    "task_identifier": task_identifier,
                    "name": "flush_period",
                    "value": self.flush_period,
                }
            )

        inputs.append(
            {
                "task_identifier": task_identifier,
                "name": "monitor_name",
                "value": self.monitor_name,
            },
        )

        return inputs

    def get_save_ascii_inputs(self, scan, lima_name):
        filename = self.get_filename(scan)
        scan_nb = scan.scan_info.get("scan_nb")
        root = self.scan_processed_directory(scan)
        stem = os.path.splitext(os.path.basename(filename))[0]
        basename = f"{stem}_{scan_nb}_{lima_name}_integrated.dat"

        return [
            {
                "task_identifier": "SaveAsciiPattern1D",
                "name": "filename",
                "value": os.path.join(root, basename),
            },
        ]

    def get_save_inputs(self, scan, lima_name, task_identifier):
        inputs = super().get_save_inputs(scan, lima_name, task_identifier)
        inputs += [
            {
                "task_identifier": task_identifier,
                "name": "nxprocess_name",
                "value": f"{lima_name}_integrate",
            }
        ]

        return inputs



from gevent import monkey

monkey.patch_all()

import logging  # noqa E402
import json  # noqa E402
from typing import Mapping, Optional  # noqa E402

from blissdata.data.node import get_session_node  # noqa E402
from ewoksjob.client import submit  # noqa E402

logger = logging.getLogger(__name__)


def session_watcher(session_name, **kw):
    session = get_session_node(session_name)
    logger.info(f"Started listening to Bliss session '{session_name}'")
    scan_types = ("scan", "scan_group")

    for ev in session.walk_on_new_events(exclude_children=scan_types, wait=True):
        if ev.type == ev.type.NEW_NODE and ev.node.type == "scan":
            info = ev.node.info
            workflows = info.get("workflows")
            if not workflows:
                continue
            filename = info.get("filename")
            scan_nb = info.get("scan_nb")
            for wfname, nxprocess in workflows.items():
                if not isinstance(nxprocess, Mapping):
                    continue
                try:
                    job_id = submit_scan_workflow(
                        **json.loads(nxprocess["configuration"]["data"])
                    )
                except Exception:
                    logger.exception(
                        f"Error when submitting workflow '{wfname}' for scan {scan_nb} of file '{filename}'"
                    )
                else:
                    if job_id is not None:
                        logger.info(
                            f"Submitted workflow '{wfname}' (JOB ID {job_id}) for scan {scan_nb} of file '{filename}'"
                        )


def submit_scan_workflow(workflow=None, **kwargs) -> Optional[str]:
    if not workflow:
        return
    future = submit(args=(workflow,), kwargs=kwargs)
    return future.task_id


def main(args) -> None:
    session_watcher(args.session)



"""API for XRPD plotting use by the processing"""

import time
from typing import Optional, Union, Tuple

import numpy

try:
    import gevent
except ImportError:
    gevent = None
try:
    from bliss import current_session
except ImportError:
    current_session = None

from silx.io import h5py_utils
from silx.utils.retry import RetryError

from ..flint import WithFlintAccess
from .plots import XrpdCurvePlot
from .plots import XrpdImagePlot
from .data import get_redis_store
from .data import add_plot
from .data import add_data
from .data import remove_plots


class Plotter(WithFlintAccess):
    def __init__(self, number_of_scans: int = 0) -> None:
        if current_session is None:
            raise ModuleNotFoundError("No module named 'bliss'")
        super().__init__()
        self._tasks = list()
        self._redis_store = get_redis_store(
            f"blissoda:{current_session.name}:{self.__class__.__name__}"
        )
        self.number_of_scans = number_of_scans

    @property
    def number_of_scans(self):
        return self._plot_scans

    @number_of_scans.setter
    def number_of_scans(self, value):
        self._plot_scans = max(value, 0)
        self._remove_old_plots()

    def add_plot(
        self,
        scan_name: str,
        lima_name: str,
        xlabel: str,
        xvalues: numpy.ndarray,
        signal_ndim: int,
        hdf5_url: Optional[str] = None,
    ) -> Tuple[str, dict]:
        plot_key, plot_info = add_plot(
            self._redis_store,
            scan_name,
            lima_name,
            signal_ndim,
            xlabel,
            xvalues,
            hdf5_url=hdf5_url,
        )
        self._remove_old_plots()
        return plot_key, plot_info

    def add_data(
        self, plot_key: str, plot_info: dict, points: numpy.ndarray, retry_options: dict
    ) -> str:
        add_data(plot_key, plot_info, points)
        self._update_plot(plot_key, plot_info, retry_options)

    def replot(self, **retry_options) -> None:
        for plot_key, plot_info in self._redis_store.items():
            self._spawn(self._update_plot, plot_key, plot_info, retry_options)

    def _update_plot(self, plot_key: str, plot_info: dict, retry_options: dict) -> None:
        if plot_info["signal_ndim"] > 1:
            imageplot = self._get_plot(plot_info["lima_name"])
            imageplot.update_plot(self._redis_store.name, plot_key, retry_options)
        curveplot = self._get_plot()
        curveplot.update_plot(self._redis_store.name, plot_key, retry_options)

    def _remove_plot(self, plot_key: str, plot_info: dict) -> None:
        if plot_info["signal_ndim"] > 1:
            imageplot = self._get_plot(plot_info["lima_name"])
            imageplot.remove_plot(self._redis_store.name, plot_key)
        curveplot = self._get_plot()
        curveplot.remove_plot(self._redis_store.name, plot_key)

    def _remove_plots(self, max_len: int):
        remove_plots(self._redis_store, max_len, self._remove_plot)

    def _remove_old_plots(self):
        self._remove_plots(self._plot_scans)

    def clear(self):
        self._remove_plots(0)

    def _get_plot(
        self, lima_name: Optional[str] = None
    ) -> Union[XrpdCurvePlot, XrpdImagePlot]:
        if lima_name:
            return super()._get_plot(f"Integrated {lima_name}", XrpdImagePlot)
        return super()._get_plot("Integrated (Last)", XrpdCurvePlot)

    def handle_workflow_result(
        self,
        future,
        lima_name: str,
        scan_name: str,
        output_url: Optional[str] = None,
        retry_timeout: int = 60,
        retry_period: int = 1,
    ):
        """Handle workflow results in the background"""
        if output_url:
            func = self.update_plot_from_file
        else:
            func = self.update_plot_from_memory
        self._spawn(
            func,
            future,
            scan_name,
            lima_name,
            output_url=output_url,
            retry_timeout=retry_timeout,
            retry_period=retry_period,
        )

    def _spawn(self, *args, **kw):
        task = gevent.spawn(*args, **kw)
        self._tasks.append(task)
        self.purge_tasks()

    def purge_tasks(self) -> int:
        self._tasks = [t for t in self._tasks if t]
        return len(self._tasks)

    def kill_tasks(self) -> int:
        gevent.killall(self._tasks)
        return self.purge_tasks()

    def update_plot_from_memory(
        self,
        future,
        scan_name: str,
        lima_name: str,
        retry_timout: int = 600,
        **retry_options,
    ) -> None:
        result = future.get(retry_timout)
        if result["radial_units"] is None:
            return

        points = result["intensity"]
        points = numpy.squeeze(points)
        xvalues = result["radial"]
        xname, xunits = result["radial_units"].split("_")
        xlabel = f"{xname} ({xunits})"

        plot_key, plot_info = self.add_plot(
            scan_name, lima_name, xlabel, xvalues, points.ndim
        )
        self.add_data(plot_key, plot_info, points, retry_options)

    def update_plot_from_file(
        self,
        future,
        scan_name: str,
        lima_name: str,
        output_url: str,
        **retry_options,
    ) -> None:
        plot_key, plot_info = self._add_plot_from_file(
            scan_name, lima_name, output_url, **retry_options
        )
        retry_period = retry_options.get("retry_period")
        while not future.ready():
            self._update_plot(plot_key, plot_info, retry_options)
            if retry_period:
                time.sleep(retry_period)
        self._update_plot(plot_key, plot_info, retry_options)

    @h5py_utils.retry()
    def _add_plot_from_file(
        self, scan_name: str, lima_name: str, output_url: str
    ) -> Tuple[str, str]:
        filename, nxdata_name = output_url.split("::")
        with h5py_utils.File(filename, mode="r") as f:
            try:
                nxdata = f[nxdata_name]
                xname = nxdata.attrs["axes"][-1]
                dset = nxdata[xname]
                xunits = dset.attrs["units"]
            except KeyError as e:
                raise RetryError(str(e))
            xlabel = f"{xname} ({xunits})"
            xvalues = dset[()]
            signal = nxdata[nxdata.attrs["signal"]]
            hdf5_url = f"{filename}::{signal.name}"
            plot_key, plot_info = self.add_plot(
                scan_name,
                lima_name,
                xlabel,
                xvalues,
                signal.ndim,
                hdf5_url=hdf5_url,
            )
            return plot_key, plot_info



"""Flint-side for Flint XRPD plots"""
from silx.gui.plot import Plot1D
from silx.gui.plot import Plot2D

try:
    from blissdata import settings
except ImportError:
    try:
        from bliss.config import settings
    except ImportError:
        settings = None

from .data import get_redis_store
from .data import get_curve_data
from .data import get_image_data
from ..flint import capture_errors


import logging

logger = logging.getLogger(__name__)


class _WithRedisData:
    def __init__(self) -> None:
        super().__init__()
        self._redis_store = None

    def _get_redis_store(self, redis_store: str) -> settings.OrderedHashObjSetting:
        if self._redis_store is None:
            self._redis_store = get_redis_store(redis_store)
        return self._redis_store

    def _get_legend(self, plot_info: dict) -> str:
        return f"{plot_info['scan_name']} ({plot_info['lima_name']})"


class XrpdCurveWidget(Plot1D, _WithRedisData):
    @capture_errors
    def remove_plot(self, redis_store: str, plot_key: str) -> None:
        plot_info = self._get_redis_store(redis_store)[plot_key]
        legend = self._get_legend(plot_info)
        logger.debug(f"remove {legend}")
        self.remove(legend=legend)

    @capture_errors
    def update_plot(self, redis_store: str, plot_key: str, retry_options: dict) -> None:
        x, y, plot_info = get_curve_data(
            self._get_redis_store(redis_store), plot_key, **retry_options
        )
        if y is None:
            return
        legend = self._get_legend(plot_info)
        logger.debug("XRPD curve plot %s", legend)
        self.addCurve(
            x, y, legend=legend, xlabel=plot_info["xlabel"], ylabel="Intensity"
        )


class XrpdImageWidget(Plot2D, _WithRedisData):
    @capture_errors
    def remove_plot(self, redis_store: str, plot_key: str) -> None:
        self.clear()

    @capture_errors
    def update_plot(self, redis_store: str, plot_key: str, retry_options: dict) -> None:
        img = self.getImage(legend=plot_key)
        if img is None:
            data = None
        else:
            data = img.getData(copy=False)
        x, y, plot_info = get_image_data(
            self._get_redis_store(redis_store), plot_key, data, **retry_options
        )
        if y is None:
            return
        origin = x[0], 0
        scale = x[1] - x[0], 1
        self.clear()
        title = self._get_legend(plot_info)
        self.setGraphTitle(title)
        logger.debug("XRPD image plot %s: %s points", title, len(y))
        self.addImage(
            y,
            legend=title,
            xlabel=plot_info["xlabel"],
            ylabel="Scan points",
            origin=origin,
            scale=scale,
        )



"""Bliss-side for Flint XRPD plots"""
try:
    from bliss.flint.client.plots import BasePlot
except ImportError:
    BasePlot = object


class XrpdCurvePlot(BasePlot):
    WIDGET = "blissoda.xrpd.widgets.XrpdCurveWidget"

    def remove_plot(self, redis_store: str, plot_key: str) -> None:
        self.submit("remove_plot", redis_store, plot_key)

    def update_plot(self, redis_store: str, plot_key: str, hdf5_options: dict) -> None:
        self.submit("update_plot", redis_store, plot_key, hdf5_options)


class XrpdImagePlot(BasePlot):
    WIDGET = "blissoda.xrpd.widgets.XrpdImageWidget"

    def remove_plot(self, redis_store: str, plot_key: str) -> None:
        self.submit("remove_plot", redis_store, plot_key)

    def update_plot(self, redis_store: str, plot_key: str, hdf5_options: dict) -> None:
        self.submit("update_plot", redis_store, plot_key, hdf5_options)



from typing import Optional
import json


def read_config(filename: Optional[str]) -> dict:
    if not filename:
        return dict()
    if filename.endswith(".json"):
        return _read_json(filename)
    else:
        return _read_poni(filename)


def _read_json(filename: str) -> dict:
    with open(filename, "r") as fp:
        return json.load(fp)


def _read_poni(filename: str) -> dict:
    options = dict()
    with open(filename, "r") as f:
        for line in f:
            if not line or line.startswith("#"):
                continue
            key, sep, value = line.partition(":")
            if not sep or not value:
                continue
            try:
                value = json.loads(value.strip())
            except json.JSONDecodeError:
                value = value.strip()
            except Exception:
                continue
            options[_parse_poni_key(key)] = value
    return options


def _parse_poni_key(k: str) -> str:
    new_key = k.strip().lower()
    if new_key == "distance":
        return "dist"
    return new_key



import os
from contextlib import contextmanager

try:
    from bliss import setup_globals
    from bliss.common.logtools import user_info
except ImportError:
    setup_globals = None
    user_info = print

from ..streamline.scanner import StreamlineScanner
from .optimize_exposure import optimize_exposure
from ..persistent import ParameterInfo


class Id31StreamlineScanner(
    StreamlineScanner,
    parameters=[
        ParameterInfo("optimize_exposure", category="exposure/attenuator"),
        ParameterInfo("optimize_pixel_value", category="exposure/attenuator"),
        ParameterInfo("optimize_nb_frames", category="exposure/attenuator"),
        ParameterInfo("optimize_max_exposure_time", category="exposure/attenuator"),
        ParameterInfo("attenuator_name", category="names"),
        ParameterInfo("newflat", category="Flat-field"),
        ParameterInfo("oldflat", category="Flat-field"),
    ],
):
    def __init__(self, **defaults):
        defaults.setdefault("detector_name", "p3")
        defaults.setdefault("attenuator_name", "atten")
        defaults.setdefault("sample_changer_name", "streamline_sc")
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_ocl_gpu",
                "integrator_name": "sigma_clip_ng",
                "extra_options": {"max_iter": 3, "thres": 0},
                "error_model": "azimuthal",  # hybrid gives weird results
                "nbpt_rad": 4096,
                "unit": "q_nm^-1",
            },
        )
        defaults.setdefault("optimize_exposure", False)
        defaults.setdefault("optimize_pixel_value", 1e5)
        defaults.setdefault("optimize_nb_frames", 3)
        defaults.setdefault("optimize_max_exposure_time", 4)
        super().__init__(**defaults)

    def _get_scan_metadata(self) -> dict:
        return dict()

    @contextmanager
    def run_context(self):
        setup_globals.shopen()
        with super().run_context():
            yield

    def measure_sample(self, count_time: float = 1, *args, **kwargs):
        att_value = None
        if self.optimize_exposure:
            detector = getattr(setup_globals, self.detector_name)
            attenuator = getattr(setup_globals, self.attenuator_name)
            att_value = attenuator.bits
            count_time = optimize_exposure(
                detector,
                tframe=0.2,
                default_att_position=None,
                desired_counts=self.optimize_pixel_value,
                dynamic_range=1 << 20,
                min_counts_per_frame=100,
                nframes_measure=1,
                nframes_default=self.optimize_nb_frames,
                reduce_desired_deviation=True,
                expose_with_integral_frames=False,
            )
            count_time = min(count_time, self.optimize_max_exposure_time)
        try:
            return super().measure_sample(count_time, *args, **kwargs)
        finally:
            if att_value is not None:
                setup_globals.att(att_value)

    def init_workflow(
        self, with_calibration: bool = False, flatfield: bool = True
    ) -> None:
        wd = "/users/opid31/ewoks/resources/workflows"
        if flatfield:
            if with_calibration:
                self.workflow = os.path.join(wd, "streamline_with_calib_with_flat.json")
            else:
                self.workflow = os.path.join(
                    wd, "streamline_without_calib_with_flat.json"
                )
        else:
            if with_calibration:
                self.workflow = os.path.join(wd, "streamline_with_calib.json")
            else:
                self.workflow = os.path.join(wd, "streamline_without_calib.json")

        user_info(f"Active data processing workflow: {self.workflow}")

    def _job_arguments(self, scan_info, processed_metadata: dict):
        args, kwargs = super()._job_arguments(scan_info, processed_metadata)
        inputs = kwargs["inputs"]
        inputs.append(
            {
                "task_identifier": "FlatFieldFromEnergy",
                "name": "newflat",
                "value": self.newflat,
            }
        )
        inputs.append(
            {
                "task_identifier": "FlatFieldFromEnergy",
                "name": "oldflat",
                "value": self.oldflat,
            }
        )
        energy = getattr(setup_globals, self.energy_name).position
        inputs.append(
            {
                "task_identifier": "FlatFieldFromEnergy",
                "name": "energy",
                "value": energy,
            }
        )
        return args, kwargs



import numpy
import pytest

from .mock_id31 import mock_id31
from .mock_id31 import setup_globals
from ..id31.optimize_exposure import optimize_exposure


@pytest.fixture
def current_session_mock():
    with mock_id31():
        yield


def test_ct(current_session_mock):
    setup_globals.p3.noise = False
    setup_globals.source.rate = 1e7
    setup_globals.sample.oyield = 0.1
    expected = 1e6

    assert setup_globals.sample.rate == expected
    setup_globals.p3.ct()
    assert setup_globals.p3.last_image.max() == expected

    setup_globals.att(1)
    assert setup_globals.sample.rate <= expected
    setup_globals.p3.ct()
    fmax = int(setup_globals.sample.rate + 0.5)
    assert setup_globals.p3.last_image.max() == fmax

    setup_globals.att(0)
    assert setup_globals.sample.rate == expected
    setup_globals.p3.ct()
    assert setup_globals.p3.last_image.max() == expected

    setup_globals.source.rate = 5e7 + 25
    with pytest.raises(RuntimeError):
        setup_globals.p3.ct()

    setup_globals.att(1)
    setup_globals.p3.ct()


def test_attenuator_optimization(current_session_mock):
    setup_globals.sample.oyield = 1
    setup_globals.p3.noise = False

    lst = list()
    for default_att_position in range(32):
        sublist = list()
        lst.append(sublist)
        flux = numpy.linspace(1e4, 1e6, 10)
        for fl in flux:
            setup_globals.source.rate = fl
            setup_globals.att(0)
            assert setup_globals.p3.rate == fl
            expo_time = optimize_exposure(
                setup_globals.p3,
                default_att_position=default_att_position,
                desired_counts=1e5,
            )
            setup_globals.ct(expo_time, setup_globals.p3)
            sublist.append(setup_globals.atten.bits)

    for sublist in lst:
        assert sublist == lst[0]



import numpy


def energy_wavelength(x):
    """keV to m and vice versa"""
    return 12.398419843320026 * 1e-10 / x


DEFAULT_CALIB = {
    "dist": 5e-2,  # 5 cm
    "poni1": 10e-2,  # 10 cm
    "poni2": 10e-2,  # 10 cm
    "rot1": numpy.radians(10),  # 10 deg
    "rot2": 0,  # 0 deg
    "rot3": 0,  # 0 deg
    "energy": 12,  # 12 keV
    "detector": "Pilatus1M",
}



import numpy

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

from . import setup_globals
from . import mock_id31
from ...id31.optimize_exposure import optimize_exposure


def simulate(
    fmin: float = 1e4,
    fmax: float = 1e6,
    default_att_position: int = 2,
    desired_counts: float = 1e5,
    energy: float = 70,
    noise: bool = False,
    reduce_desired_deviation: bool = True,
    expose_with_integral_frames: bool = False,
    max_expo_time: float = 4,
    nframes_default: int = 3,
):
    setup_globals.sample.oyield = 1
    setup_globals.p3.noise = noise
    setup_globals.energy.position = energy

    lst = list()
    flux = numpy.linspace(fmin, fmax, 500)
    for fl in flux:
        setup_globals.source.rate = fl
        setup_globals.att(0)
        assert setup_globals.p3.rate == fl
        expo_time = optimize_exposure(
            setup_globals.p3,
            nframes_default=nframes_default,
            default_att_position=default_att_position,
            desired_counts=desired_counts,
            reduce_desired_deviation=reduce_desired_deviation,
            expose_with_integral_frames=expose_with_integral_frames,
        )
        if max_expo_time:
            expo_time = min(expo_time, max_expo_time)
        setup_globals.ct(expo_time, setup_globals.p3)
        lst.append(
            (setup_globals.p3.last_image.max(), expo_time, setup_globals.atten.bits)
        )
    max_intensity, expo_time, attenuator = zip(*lst)

    fig, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(8, 10))

    ax1.plot(flux, max_intensity)
    # ax1.set_xlabel("Pixel Value (Hz)")
    ax1.set_ylabel("Pixel value (counts)")

    xmin = flux[0]
    xmax = flux[0] + 0.85 * (flux[-1] - flux[0])
    ax1.hlines(desired_counts, xmin, xmax, linestyles="dotted")
    ax1.text(xmax, desired_counts, "target value", ha="left", va="center")

    ax2.plot(flux, expo_time)
    # ax2.set_xlabel("Pixel Value (Hz)")
    ax2.set_ylabel("Exposure time (s)")

    xmin = flux[0]
    xmax = flux[0] + 0.85 * (flux[-1] - flux[0])
    expo_default = nframes_default * 0.2
    ax2.hlines(expo_default, xmin, xmax, linestyles="dotted")
    ax2.text(xmax, expo_default, "default time", ha="left", va="center")

    ax3.plot(flux, attenuator)
    ax3.set_xlabel("Pixel Value (Hz)")
    ax3.set_ylabel("Attenuator position")

    xmin = flux[0]
    xmax = flux[0] + 0.85 * (flux[-1] - flux[0])
    ax3.hlines(default_att_position, xmin, xmax, linestyles="dotted")
    ax3.text(xmax, default_att_position, "start position", ha="left", va="center")

    fig.suptitle(
        f"{flux[0]:.0e} Hz - {flux[-1]:.0e} Hz, {setup_globals.energy.position} keV"
    )


if __name__ == "__main__":
    with mock_id31():
        simulate()
        plt.show()



from blissoda.xrpd.processor import XrpdProcessor
from blissoda.exafs.plotter import ExafsPlotter
from blissoda.streamline.scanner import StreamlineScanner

from blissoda.bm23.exafs_plotter import Bm23ExafsPlotter
from blissoda.bm02.xrpd_processor import Bm02XrpdProcessor
from blissoda.id11.xrpd_processor import Id11XrpdProcessor
from blissoda.id31.xrpd_processor import Id31XrpdProcessor
from blissoda.id31.streamline_scanner import Id31StreamlineScanner

from blissoda.demo.exafs import exafs_plotter
from blissoda.demo.xrpd import xrpd_processor
from blissoda.demo.id22 import stscan_processor
from blissoda.demo.streamline import streamline_scanner

from blissoda.demo.user_scripts.exafs import exafs_demo
from blissoda.demo.user_scripts.id22 import id22_demo
from blissoda.demo.user_scripts.streamline import streamline_demo
from blissoda.demo.user_scripts.xrpd import xrpd_demo

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None


def all_print():
    _print_objects(exafs_plotter)
    _print_objects(xrpd_processor)
    _print_objects(stscan_processor)
    _print_objects(streamline_scanner)

    _print_objects(XrpdProcessor())
    _print_objects(ExafsPlotter())
    _print_objects(StreamlineScanner())

    _print_objects(Bm23ExafsPlotter())
    _print_objects(Bm02XrpdProcessor())
    _print_objects(Id11XrpdProcessor())
    _print_objects(Id31XrpdProcessor())

    _print_objects(Id31StreamlineScanner())


def all_demo():
    print()
    print("===================")
    setup_globals.newcollection("exafs")
    exafs_demo()

    print()
    print("===================")
    setup_globals.newcollection("id22")
    id22_demo()

    print()
    print("===================")
    setup_globals.newcollection("streamline")
    streamline_demo()

    print()
    print("===================")
    setup_globals.newcollection("xrpd")
    xrpd_demo()

    print()
    print("===================")
    setup_globals.newcollection()


def _print_objects(obj):
    print()
    print("===================")
    print(obj._parameters.name)
    print(obj.__info__())



import json
from typing import Optional, List
from .calib import DEFAULT_CALIB

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None

from ..xrpd.processor import XrpdProcessor
from ..persistent import ParameterInfo


class DemoXrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("config_filename", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
    ],
):
    def __init__(self, **defaults) -> None:
        defaults.setdefault("lima_names", ["difflab6"])
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_cython",
                "nbpt_rad": 4096,
                "radial_range_min": 1,
                "unit": "q_nm^-1",
            },
        )
        super().__init__(**defaults)
        self._ensure_config_filename()

    def get_integrate_inputs(
        self, scan, lima_name: str, task_identifier: str
    ) -> List[dict]:
        self._ensure_config_filename()
        inputs = super().get_integrate_inputs(scan, lima_name, task_identifier)
        inputs.append(
            {"task_identifier": task_identifier, "name": "demo", "value": True}
        )
        return inputs

    def _ensure_config_filename(self):
        if self.config_filename:
            return
        cfgfile = "/tmp/test.json"
        poni = DEFAULT_CALIB
        with open(cfgfile, "w") as f:
            json.dump(poni, f)
        self.config_filename = cfgfile

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return self.config_filename

    def get_integration_options(self, lima_name: str) -> Optional[dict]:
        return self.integration_options.to_dict()


if setup_globals is None:
    xrpd_processor = None
else:
    try:
        xrpd_processor = DemoXrpdProcessor()
    except ImportError:
        xrpd_processor = None



"""Automatic pyfai integration for every scan with saving and plotting"""

import os
import tempfile
from glob import glob
from typing import Optional, Dict

from ..xrpd.processor import XrpdProcessor
from ..persistent import ParameterInfo
from ..utils.pyfai import read_config


class Id11XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("pyfai_config_directory", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
    ],
):
    def __init__(self, **defaults) -> None:
        defaults.setdefault("pyfai_config_directory", tempfile.gettempdir())
        super().__init__(**defaults)

    def _info_categories(self) -> Dict[str, dict]:
        categories = super()._info_categories()
        if not self.lima_names:
            return categories
        categories["PyFai"]["integration_options"] = "... (see below)"
        lima_name = self.lima_names[0]
        categories["PyFai integration"] = {
            "1. JSON file": self._detector_config_filename(lima_name, ".json"),
            "2. PONI file": self._detector_config_filename(lima_name, ".poni"),
            "3. User": self.integration_options,
            "Merged": self.get_integration_options(lima_name),
        }
        return categories

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        return None

    def get_integration_options(self, lima_name: str) -> Optional[dict]:
        options = self._default_azint_options(lima_name)
        options.update(self._default_calib_options(lima_name))
        integration_options = self.integration_options
        if integration_options:
            options.update(integration_options.to_dict())
        return options

    def _default_azint_options(self, lima_name: str) -> dict:
        filename = self._detector_config_filename(lima_name, ".json")
        return read_config(filename)

    def _default_calib_options(self, lima_name: str) -> dict:
        filename = self._detector_config_filename(lima_name, ".poni")
        return read_config(filename)

    def _detector_config_filename(self, lima_name: str, ext: str) -> Optional[str]:
        pyfai_config_directory = self.pyfai_config_directory
        if not pyfai_config_directory:
            pyfai_config_directory = tempfile.gettempdir()
        pattern = os.path.join(pyfai_config_directory, lima_name, f"*{ext}")
        files = sorted(glob(pattern))
        if not files:
            return
        return files[-1]



from pprint import pprint
from typing import List
from .calib import DEFAULT_CALIB
import json

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None
from ..id22.stscan_processor import StScanProcessor
from ..id22.xrpd_processor import Id22XrpdProcessor


class DemoStScanProcessor(StScanProcessor):
    def __init__(self) -> None:
        super().__init__(
            convert_workflow="/tmp/demo/convert.json",
            rebinsum_workflow="/tmp/demo/rebinsum.json",
            extract_workflow="/tmp/demo/extract.json",
        )

    def _submit_job(self, workflow, inputs, **kw):
        print("\nSubmit workfow")
        print(workflow)
        print("Inputs:")
        pprint(inputs)
        print("Options:")
        pprint(kw)


class DemoId22XrpdProcessor(Id22XrpdProcessor):
    def __init__(self, **defaults) -> None:
        defaults.setdefault("lima_names", ["perkinelmer"])
        defaults.setdefault(
            "integration_options",
            {
                "method": "no_csr_cython",
                "nbpt_rad": 4096,
                "radial_range_min": 1,
                "unit": "q_nm^-1",
            },
        )
        super().__init__(**defaults)
        self._ensure_config_filename()

    def get_integrate_1d_inputs(self, scan, lima_name: str) -> List[dict]:
        inputs = super().get_integrate_1d_inputs(scan, lima_name)
        inputs.append({"task_identifier": "Integrate1D", "name": "demo", "value": True})
        return inputs

    def _ensure_config_filename(self):
        if self.pyfai_config:
            return
        cfgfile = "/tmp/test.json"
        poni = DEFAULT_CALIB
        with open(cfgfile, "w") as f:
            json.dump(poni, f)
        self.pyfai_config = cfgfile


if setup_globals is None:
    stscan_processor = None
    xrpd_processor = None
else:
    stscan_processor = DemoStScanProcessor()
    xrpd_processor = DemoId22XrpdProcessor()



from blissoda.demo.id22 import stscan_processor
from blissoda.demo.id22 import xrpd_processor

try:
    from bliss import setup_globals
except ImportError:
    setup_globals = None


def id22_demo():
    stscan_processor.submit_workflows()


def id22_xrpd_demo(expo=0.2, npoints=10):
    xrpd_processor.enable(setup_globals.difflab6)
    pct(
        expo,
        setup_globals.difflab6,
        setup_globals.diode1,
        setup_globals.diode2,
    )
    setup_globals.loopscan(
        npoints,
        expo,
        setup_globals.difflab6,
        setup_globals.diode1,
        setup_globals.diode2,
    )


def pct(*args, **kw):
    s = setup_globals.ct(*args, **kw)
    xrpd_processor.on_new_scan(s)



import sys
import argparse
from typing import Optional
from . import cli_log_utils
from . import workflow_server


def create_argument_parser(shell=False):
    parser = argparse.ArgumentParser(
        description="CLI for Bliss Online Data Analysis", prog="blissoda"
    )

    subparsers = parser.add_subparsers(help="Commands", dest="command")
    workflow = subparsers.add_parser(
        "workflows",
        help="Trigger workflows (specified in scan info) for each scan that saves data",
    )
    add_workflow_parameters(workflow, shell=shell)
    return parser


def main(argv=None, shell=True) -> Optional[int]:
    parser = create_argument_parser(shell=shell)

    if argv is None:
        argv = sys.argv
    args = parser.parse_args(argv[1:])

    if args.command == "workflows":
        return command_workflow(args, shell=shell)
    else:
        parser.print_help()
        return command_default(args, shell=shell)


def command_default(args, shell: bool = False) -> Optional[int]:
    if shell:
        return 0
    else:
        return None


def command_workflow(args, shell: bool = False) -> Optional[int]:
    if shell:
        cli_log_utils.apply_log_parameters(args)
    workflow_server.main(args)
    if shell:
        return 0
    else:
        return None


def add_workflow_parameters(parser, shell: bool = False) -> None:
    if shell:
        cli_log_utils.add_log_parameters(parser)
    parser.add_argument("session", type=str, help="Bliss session")


if __name__ == "__main__":
    sys.exit(main())



from blissoda.demo.streamline import streamline_sc
from blissoda.demo.streamline import streamline_scanner


def streamline_demo(with_calibration=False):
    # Nothing loaded and 2 holders in the tray:
    streamline_scanner.eject()
    streamline_sc.fill_tray(2)

    # Initialize workflow with or without calibration
    streamline_scanner.init_workflow(with_calibration=with_calibration)
    if with_calibration:
        streamline_scanner.calib(0.1)

    # Measure all holders
    streamline_scanner.run(0.1)



import mock
import pytest
from copy import deepcopy
from collections.abc import MutableMapping
from typing import Any, Iterator, Mapping
from ..persistent import WithPersistentParameters


@pytest.fixture
def mock_persistent():
    remote_dict = MockHashObjSetting()

    def init(self, **defaults):
        for name, value in defaults.items():
            remote_dict.setdefault(name, value)
        self._parameters = remote_dict

    with mock.patch.object(WithPersistentParameters, "__init__", init):
        yield remote_dict


class MockHashObjSetting(MutableMapping):
    def __init__(self) -> None:
        self._adict = dict()
        super().__init__()

    def __repr__(self) -> str:
        return repr(self._adict)

    def get_all(self) -> dict:
        return deepcopy(self._adict)

    def __getitem__(self, key: str) -> Any:
        return self._adict[key]

    def __setitem__(self, key: str, value: Any) -> None:
        if isinstance(value, Mapping):
            value = deepcopy(value)
        self._adict[key] = value

    def __delitem__(self, key: str) -> None:
        del self._adict[key]

    def __iter__(self) -> Iterator[Any]:
        yield from self._adict

    def __len__(self) -> int:
        return len(self._adict)



from ..exafs.plotter import ExafsPlotter


class Bm23ExafsPlotter(ExafsPlotter):
    def __init__(self, **defaults) -> None:
        defaults.setdefault("workflow", "/users/opd23/ewoks/online.ows")
        defaults.setdefault("_scan_type", "cont")
        counters = defaults.setdefault("_counters", dict())
        counters.setdefault(
            "cont",
            {
                "mu_name": "mu_trans",
                "energy_name": "energy_cenc",
                "energy_unit": "keV",
            },
        )
        counters.setdefault(
            "step",
            {
                "mu_name": "mu_trans",
                "energy_name": "eneenc",
                "energy_unit": "keV",
            },
        )
        super().__init__(**defaults)

    def _scan_type_from_scan(self, scan) -> str:
        if "exafs_step" in scan.name:
            return "step"
        else:
            return "cont"



"""Bliss-side client for Flint EXAFS plots"""

try:
    from bliss.flint.client.plots import BasePlot
except ImportError:
    BasePlot = object


class ExafsPlot(BasePlot):
    WIDGET = "blissoda.exafs.widgets.ExafsWidget"

    def clear(self):
        self.submit("clear")

    def remove_scan(self, legend):
        self.submit("remove_scan", legend)

    def update_scan(self, legend, data, color=None):
        self.submit("update_scan", legend, data, color=color)

    def get_scans(self):
        return self.submit("get_scans")



import logging
import numpy
from . import attenuator

logger = logging.getLogger(__name__)


class BlObject:
    def __init__(self, parent=None):
        self.parent = parent

    def __getattr__(self, name):
        return getattr(self.parent, name)

    @property
    def oyield(self):
        """A value between 0 and 1"""
        return 1

    @property
    def rate(self):
        """Hz"""
        return self.parent.rate * self.oyield


class Source:
    def __init__(self):
        self.rate = 1e7  # Hz

    @property
    def rate(self):
        return self._rate

    @rate.setter
    def rate(self, value):
        self._rate = value


class Mono(BlObject):
    def __init__(self, **kw):
        super().__init__(**kw)
        self.energy = 70  # keV

    @property
    def position(self):
        return self.energy

    @position.setter
    def position(self, value):
        self.energy = value


class Attenuator(BlObject):
    def __init__(self, **kw):
        super().__init__(**kw)
        self.bits = 0

    @property
    def oyield(self):
        return attenuator.SiO2trans(self.energy, self.bits)


class DiffractingSample(BlObject):
    def __init__(self, **kw):
        super().__init__(**kw)
        self.oyield = 1

    @property
    def oyield(self):
        return self._oyield

    @oyield.setter
    def oyield(self, value):
        self._oyield = value


class LimaDetector(BlObject):
    def __init__(self, **kw):
        super().__init__(**kw)
        self.tframe = 0.2  # sec
        self.dynamic_range = 1 << 20
        self.damage_rate = 5e6  # Hz
        self._last_image = None
        self.noise = True
        self.shape = 11, 10

    @property
    def proxy(self):
        return self

    @property
    def last_image(self):
        return self._last_image

    def ct(self, expo_time=1):
        # Devide exposure time in frames
        nframes = expo_time / self.tframe
        expo_left = expo_time - int(nframes) * self.tframe
        nframes = int(nframes)
        frame_times = [self.tframe] * nframes + [expo_left]

        # Generate frames
        frames, fmax, rmax = zip(
            *(self._generate_frame(tframe) for tframe in frame_times)
        )

        # Sum frames
        self._last_image = sum(frames)

        # Warn about exceeding dynamic range
        fmax = sorted(set(fmax) - {None})
        if fmax:
            logger.warning(f"dynamic range exceeded ({fmax[-1]} cts/frame)")

        # Raise error when damaging the detector
        rmax = sorted(set(rmax) - {None})
        if rmax:
            raise RuntimeError(f"detector damaged at {rmax[-1]:e} Hz")

    def _generate_frame(self, tframe):
        # Generate frame
        counts = int(self.rate * tframe + 0.5)
        if self.noise:
            frame = numpy.random.poisson(counts, size=self.shape)
            fmax = frame.max()
        else:
            frame = numpy.full(self.shape, counts)
            fmax = counts
        if tframe:
            rmax = fmax / tframe
        else:
            rmax = 0
        if fmax >= self.dynamic_range:
            frame = numpy.clip(frame, 0, self.dynamic_range - 1)
        else:
            fmax = None
        if rmax <= self.damage_rate:
            rmax = None
        return frame, fmax, rmax



"""Workflow execution and Flint EXAFS plotting during a scan"""

import time
import os
import logging
import gevent
from collections import OrderedDict
from typing import List, Optional, Tuple, Dict

from silx.io.h5py_utils import top_level_names

try:
    from bliss.common.plot import get_flint
    from bliss.flint.client.proxy import FlintClient
except ImportError:
    get_flint = None
    FlintClient = None

from ewoksjob.client import submit
from ewoksjob.client import get_future

from ..persistent import WithPersistentParameters
from .plots import ExafsPlot
from ..persistent import ParameterInfo


logger = logging.getLogger(__name__)


class ExafsPlotter(
    WithPersistentParameters,
    parameters=[
        ParameterInfo("refresh_period"),
        ParameterInfo("max_scans"),
        ParameterInfo("enabled"),
        ParameterInfo("workflow"),
        ParameterInfo("_counters"),
        ParameterInfo("_scan_type"),
        ParameterInfo("_color_index"),
    ],
):
    """Run a scan, execute a workflow every x seconds during the scan
    and plot the results in Flint. A fixed number of n scans stay plotted.
    """

    def __init__(self, **defaults) -> None:
        defaults.setdefault("refresh_period", 2)  # seconds
        defaults.setdefault("max_scans", 3)
        defaults.setdefault("enabled", True)
        defaults.setdefault("_counters", dict())
        defaults.setdefault("_color_index", 0)
        super().__init__(**defaults)

        # Fixed parameters
        self._plot_id = "EXAFS"
        self._plot_names = {
            "flatten_mu": "mu",
            "chi_weighted_k": "chi",
            "ft_mag": "ft",
            "noise_savgol": "noise",
        }

        # Runtime data
        self._scans = OrderedDict()  # scan_id -> scan_info
        self._plot: Optional[ExafsPlot] = None
        self._client = None

    @property
    def counters(self) -> dict:
        return self._counters.get(self.scan_type, dict())

    @property
    def scan_type(self):
        return self._scan_type

    @scan_type.setter
    def scan_type(self, value):
        if value not in self._counters:
            raise ValueError(f"Valid scan types are: {list(self._counters)}")
        self._scan_type = value

    @property
    def mu_name(self) -> Optional[str]:
        return self.counters.get("mu_name")

    @mu_name.setter
    def mu_name(self, value):
        self.counters["mu_name"] = value

    @property
    def energy_name(self) -> Optional[str]:
        return self.counters.get("energy_name")

    @energy_name.setter
    def energy_name(self, value):
        self.counters["energy_name"] = value

    @property
    def energy_unit(self) -> Optional[str]:
        return self.counters.get("energy_unit")

    @energy_unit.setter
    def energy_unit(self, value):
        self.counters["energy_unit"] = value

    def _scan_type_from_scan(self, scan) -> str:
        raise NotImplementedError

    def run(self, scan):
        if not self.enabled:
            scan.run()
            return

        self.scan_type = self._scan_type_from_scan(scan)

        filename = scan.writer.filename
        if os.path.exists(filename):
            scannr = max(int(float(s)) for s in top_level_names(filename)) + 1
        else:
            scannr = 1
        scan_legend = f"{scannr}.1"
        scan_url = f"silx://{filename}::/{scannr}.1"
        scan_id = scan_url
        scan_name = os.path.split(os.path.dirname(filename))[-1]
        scan_name = f"{scan_name}:{scannr}.1"
        scan_color = self._COLOR_PALETTE[min(self._color_index, self._n_colors - 1)]
        self._color_index = (self._color_index + 1) % self._n_colors
        args = scan_id, scan_name, scan_legend, scan_color, scan_url

        update_loop = gevent.spawn(self._plotting_loop, *args)

        try:
            scan.run()
        finally:
            try:
                try:
                    if not update_loop:
                        update_loop.get()
                    update_loop.kill()
                    self._submit_and_plot(scan_id)
                finally:
                    self._removed_failed_processing()
                    self._purge_plots()
            except Exception:
                logger.warning("Post-scan update failed", exc_info=True)

    def clear(self):
        """Remove all scan curves in all plots"""
        self._get_plot().clear()

    def refresh(self):
        """Refresh all plots with the current processed data"""
        for scan_id in self._scans:
            self._update_scan_plot(scan_id)

    def reprocess(self):
        """Reprocess all scans and update all curves"""
        for scan_id in self._scans:
            self._submit_and_plot(scan_id)

    def dump(self) -> List[Tuple[str, dict]]:
        skip = "future", "result", "previous_result"
        return [
            (scan_id, {k: v for k, v in info.items() if k not in skip})
            for scan_id, info in self._scans.items()
        ]

    def load(self, data: List[Tuple[str, dict]]) -> None:
        for scan_id, scan_info in data:
            self._init_scan_cache(scan_id, **scan_info)
        self.refresh()

    def _submit_and_plot(self, scan_id: str):
        self._submit_workfow(scan_id)
        self._update_scan_plot(scan_id)

    def _plotting_loop(
        self,
        scan_id: str,
        scan_name: str,
        scan_legend: str,
        scan_color: tuple,
        scan_url: str,
    ):
        self._init_scan_cache(
            scan_id,
            scan_name=scan_name,
            scan_legend=scan_legend,
            scan_color=scan_color,
            scan_url=scan_url,
        )
        t0 = time.time()
        while True:
            t1 = time.time()
            sleep_time = max(t0 + self.refresh_period - t1, 0)
            gevent.sleep(sleep_time)
            t0 = t1
            try:
                self._submit_and_plot(scan_id)
            except Exception as e:
                logger.error(f"EXAFS workflow or plot failed ({e})")

    def _init_scan_cache(self, scan_id: str, **kw):
        scan_info = self._scans.get(scan_id)
        if scan_info is not None:
            return
        scan_info = {
            "job_id": None,
            "future": None,
            "result": None,
            "previous_result": None,
            "scan_legend": None,
            "scan_url": None,
            "scan_name": None,
            "scan_color": None,
        }
        scan_info.update(kw)
        self._scans[scan_id] = scan_info
        return scan_info

    def _purge_plots(self, max_scans=None):
        if max_scans is None:
            max_scans = self.max_scans
        npop = max(len(self._scans) - max_scans, 0)
        for _ in range(npop):
            _, scan_info = self._scans.popitem(last=False)
            self._remove_scan_plot(scan_info["scan_legend"])

    def remove_scan(self, scan_legend: str):
        scans = OrderedDict()
        for scan_id, scan_info in self._scans.items():
            if scan_info["scan_legend"] == scan_legend:
                self._remove_scan_plot(scan_legend)
            else:
                scans[scan_id] = scan_info
        self._scans = scans

    def _removed_failed_processing(self):
        scans = OrderedDict()
        for scan_id, scan_info in self._scans.items():
            if scan_info["previous_result"]:
                scans[scan_id] = scan_info
            else:
                self._remove_scan_plot(scan_info["scan_legend"])
        self._scans = scans

    def _remove_scan_plot(self, legend: str):
        self._get_plot().remove_scan(legend)

    def _update_scan_plot(self, scan_id: str):
        """Update all scan curves in all plots"""
        if self._plot_exists():
            # Update the existing plot for the requested scan
            self._update_scan(scan_id)
        else:
            # Create a fresh plot with all scan curves
            for _scan_id in self._scans:
                self._update_scan(_scan_id)

    def _update_scan(self, scan_id: str):
        """Update the scan curve in all plots"""
        result, scan_info = self._get_data(scan_id)
        if result is None:
            return
        data = {self._plot_names[k]: v for k, v in result.items()}
        self._get_plot().update_scan(
            scan_info["scan_legend"], data, color=scan_info["scan_color"]
        )

    def _get_plot(self) -> ExafsPlot:
        """Launches Flint and creates the plot when either is missing"""
        if not self._plot_exists():
            self._plot = self._get_flint().get_plot(
                ExafsPlot, unique_name=self._plot_id
            )
        return self._plot

    def _plot_exists(self) -> bool:
        if self._plot is None:
            return False
        return self._get_flint().is_plot_exists(self._plot_id)

    def _get_flint(self) -> FlintClient:
        """Launches Flint when missing"""
        if not self._flint_exists():
            self._client = get_flint()
        return self._client

    def _flint_exists(self):
        if self._client is None:
            return False
        try:
            if self._client.is_available():
                return True
        except FileNotFoundError:
            pass
        return False

    def _get_data(self, scan_id: str) -> Optional[Tuple[dict, dict]]:
        """Get data and curve legend for a scan when available.
        Blocks when the workflow for the scan is still running"""
        scan_info = self._scans.get(scan_id)
        if scan_info is None:
            return scan_info["previous_result"], scan_info
        result = scan_info.get("result")
        if result is not None:
            return result, scan_info
        future = scan_info.get("future")
        if future is None:
            job_id = scan_info.get("job_id")
            if job_id is None:
                return scan_info["previous_result"], scan_info
            future = scan_info["future"] = get_future(job_id)
        try:
            result = future.get()
        except Exception:
            return scan_info["previous_result"], scan_info
        result = result["plot_data"]

        scan_info["previous_result"] = result
        scan_info["result"] = result
        return result, scan_info

    def _submit_workfow(self, scan_id: str) -> None:
        """Submit the data processing for a scan"""
        scan_info = self._scans.get(scan_id)
        if not scan_info or not scan_info["scan_url"]:
            return
        inputs = list()
        input_information = {
            "channel_url": f"{scan_info['scan_url']}/measurement/{self.energy_name}",
            "spectra_url": f"{scan_info['scan_url']}/measurement/{self.mu_name}",
            "energy_unit": self.energy_unit,
        }
        inputs.append(
            {
                "label": "xas input",
                "name": "input_information",
                "value": input_information,
            }
        )
        inputs.append(
            {
                "label": "plotdata",
                "name": "plot_names",
                "value": list(self._plot_names),
            }
        )
        future = submit(args=(self.workflow,), kwargs={"inputs": inputs})
        scan_info["future"] = future
        scan_info["job_id"] = future.task_id
        scan_info["result"] = None

    def _info_categories(self) -> Dict[str, dict]:
        return {"parameters": self._param_info(), "processing": self._process_info()}

    def _process_info(self) -> str:
        return {
            scan_info[
                "scan_name"
            ]: f"Job {scan_info['job_id']}, Processed = {bool(scan_info['previous_result'])}"
            for scan_info in self._scans.values()
        }

    def _param_info(self) -> dict:
        return {
            "enabled": self.enabled,
            "scan_type": self.scan_type,
            "workflow": self.workflow,
            "mu": self.mu_name,
            "energy": self.energy_name,
            "energy_unit": self.energy_unit,
            "refresh_period": self.refresh_period,
            "max_scans": self.max_scans,
        }

    @property
    def _n_colors(self):
        return min(len(self._COLOR_PALETTE), self.max_scans + 1)

    _COLOR_PALETTE = [
        (87, 81, 212),
        (235, 171, 33),
        (176, 69, 0),
        (0, 197, 248),
        (207, 97, 230),
        (0, 166, 107),
        (184, 0, 87),
        (0, 138, 248),
        (0, 110, 0),
        (0, 186, 171),
        (255, 145, 133),
        (133, 133, 0),
    ]



"""Automatic pyfai integration for every scan with saving and plotting"""

from typing import Optional
from ..xrpd.processor import XrpdProcessor
from ..persistent import ParameterInfo


class Bm02XrpdProcessor(
    XrpdProcessor,
    parameters=[
        ParameterInfo("config_filename", category="PyFai"),
        ParameterInfo("integration_options", category="PyFai"),
    ],
):
    def __init__(self, **defaults) -> None:
        defaults.setdefault("config_filename", {"WOS": "", "D5": ""})
        defaults.setdefault(
            "integration_options",
            {
                "WOS": {
                    "method": "no_csr_cython",
                    "nbpt_rad": 4096,
                    "unit": "q_nm^-1",
                },
                "D5": {
                    "method": "no_csr_cython",
                    "nbpt_rad": 4096,
                    "unit": "q_nm^-1",
                },
            },
        )
        super().__init__(**defaults)

    def get_config_filename(self, lima_name: str) -> Optional[str]:
        try:
            return self.config_filename[lima_name]
        except KeyError:
            raise RuntimeError(
                f"Missing pyfai configuration file (poni or json) for '{lima_name}'"
            ) from None

    def get_integration_options(self, lima_name: str) -> Optional[dict]:
        try:
            return self.integration_options[lima_name]
        except KeyError:
            raise RuntimeError(
                f"Missing pyfai integration options for '{lima_name}'"
            ) from None



import sys
import logging

LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}


def add_log_parameters(parser):
    parser.add_argument(
        "-l",
        "--log",
        type=str.lower,
        choices=list(LEVELS),
        default="warning",
        help="Log level",
    )


def apply_log_parameters(args):
    logger = logging.getLogger()
    level = LEVELS[args.log]
    logger.setLevel(level)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)-8s - %(message)s"
    )

    class StdOutFilter(logging.Filter):
        def filter(self, record):
            return record.levelno < logging.WARNING

    class StdErrFilter(logging.Filter):
        def filter(self, record):
            return record.levelno >= logging.WARNING

    if level < logging.WARNING:
        h = logging.StreamHandler(sys.stdout)
        h.addFilter(StdOutFilter())
        h.setLevel(level)
        if formatter is not None:
            h.setFormatter(formatter)
        logger.addHandler(h)

    h = logging.StreamHandler(sys.stderr)
    h.addFilter(StdErrFilter())
    h.setLevel(level)
    if formatter is not None:
        h.setFormatter(formatter)
    logger.addHandler(h)



import logging
import gevent.lock
from functools import wraps
from typing import Type, Callable

try:
    from bliss.common.plot import get_flint
    from bliss.flint.client.proxy import FlintClient
    from bliss.flint.client.plots import BasePlot
except ImportError:
    get_flint = None
    FlintClient = None
    BasePlot = object

logger = logging.getLogger(__name__)


class WithFlintAccess:
    def __init__(self) -> None:
        self._client = None
        self._clientlock = gevent.lock.RLock()
        self._plots = dict()

    def _get_plot(self, plot_id: str, plot_cls: Type[BasePlot]) -> BasePlot:
        """Launches Flint and creates the plot when either is missing"""
        plot = self._plots.get(plot_id)
        if plot is None:
            plot = self._flint_client.get_plot(plot_cls, unique_name=plot_id)
            self._plots["plot_id"] = plot
        return plot

    @property
    def _flint_client(self) -> FlintClient:
        """Launches Flint when missing"""
        with self._clientlock:
            try:
                if self._client.is_available():
                    return self._client
            except (FileNotFoundError, AttributeError):
                pass
            self._client = get_flint()
            self._plots = dict()
            return self._client


def capture_errors(method) -> Callable:
    @wraps(method)
    def wrapper(*args, **kw):
        try:
            return method(*args, **kw)
        except Exception as e:
            msg = f"Flint plot error: {e}"
            logger.error(msg, exc_info=True)

    return wrapper



from typing import Tuple
import os


def get_dataset_dir(dataset_filename: str) -> str:
    return os.path.dirname(os.path.abspath(dataset_filename))


def get_collection_dir(dataset_filename: str) -> str:
    return _abs_join(get_dataset_dir(dataset_filename), "..")


def get_raw_dir(dataset_filename: str) -> str:
    return _abs_join(get_collection_dir(dataset_filename), "..")


def get_proposal_dir(dataset_filename: str) -> str:
    dirname = get_raw_dir(dataset_filename)
    if os.path.basename(dirname) == "raw":
        # version 2
        return _abs_join(dirname, "..")
    # version 1: proposal == raw
    return dirname


def get_processed_dir(dataset_filename: str) -> str:
    version, dirname = get_directory_version(dataset_filename)
    if version == 3:
        return _abs_join(dirname, "..", "PROCESSED_DATA")
    if version == 2:
        return _abs_join(dirname, "..", "processed")
    return _abs_join(dirname, "processed")


def get_workflows_dir(dataset_filename: str) -> str:
    return os.path.join(get_processed_dir(dataset_filename), "workflows")


def get_nobackup_dir(dataset_filename: str) -> str:
    version, dirname = get_directory_version(dataset_filename)
    if version == 3:
        return _abs_join(dirname, "..", "NOBACKUP")
    if version == 2:
        return _abs_join(dirname, "..", "_nobackup")
    return _abs_join(dirname, "_nobackup")


def get_directory_version(dataset_filename: str) -> Tuple[int, str]:
    """Returns directory structure version number and the raw data directory"""
    dirname = get_raw_dir(dataset_filename)
    if os.path.basename(dirname) == "RAW_DATA":
        return 3, dirname
    if os.path.basename(dirname) == "raw":
        return 2, dirname
    # proposal == raw
    return 1, dirname


def get_session_dir(dataset_filename: str) -> str:
    return _abs_join(get_proposal_dir(dataset_filename), "..")


def _abs_join(*args):
    return os.path.abspath(os.path.join(*args))



from collections.abc import MutableMapping
from dataclasses import dataclass
from typing import Callable, Any, Iterator, Mapping, Iterable, Dict, Optional

try:
    from bliss import current_session
except ImportError:
    current_session = None
try:
    from bliss.common.utils import autocomplete_property
except ImportError:
    autocomplete_property = property
try:
    from blissdata import settings
except ImportError:
    try:
        from bliss.config import settings
    except ImportError:
        settings = None
from .utils.info import format_info


@dataclass
class ParameterInfo:
    name: str
    category: str = "parameters"
    hidden: Optional[bool] = None

    def __post_init__(self):
        if self.hidden is None:
            self.hidden = self.name.startswith("_")


class WithPersistentParameters:
    """Adds parameters as properties that will be stored in Redis

    .. code:: python

        class MyClass(WithPersistentParameters, parameters=["a", "b"])
            pass

        myobj = MyClass()
        myobj.a = 10
        myobj.b = None  # remove
    """

    _PARAMETERS: Dict[str, ParameterInfo] = dict()

    def __init__(self, **defaults) -> None:
        if current_session is None:
            raise ModuleNotFoundError("No module named 'bliss'")
        self._parameters = settings.HashObjSetting(
            f"blissoda:{current_session.name}:{self.__class__.__name__}"
        )
        for name, value in defaults.items():
            if self._get_parameter(name) is None:
                self._set_parameter(name, value)

    def __init_subclass__(cls, parameters=None, **kw) -> None:
        super().__init_subclass__(**kw)
        if parameters:
            extra_parameters = list()
            for p in parameters:
                if isinstance(p, str):
                    p = ParameterInfo(name=p)
                elif isinstance(p, Mapping):
                    p = ParameterInfo(**p)
                extra_parameters.append(p)
            extra_parameters = {p.name: p for p in extra_parameters}
        else:
            extra_parameters = dict()
        cls._PARAMETERS = {**cls._PARAMETERS, **extra_parameters}
        for name in extra_parameters:
            add_parameter_property(cls, name)

    def __dir__(self) -> Iterable[str]:
        return list(super().__dir__()) + [
            name for name, param in self._PARAMETERS.items() if not param.hidden
        ]

    def _info_categories(self) -> Dict[str, dict]:
        categories = dict()
        all_values = self._parameters.get_all()
        for name, param in self._PARAMETERS.items():
            if param.hidden:
                continue
            category = categories.setdefault(param.category, dict())
            category[name] = all_values.get(name)
        return categories

    def __info__(self) -> str:
        return "\n\n".join(
            [
                f"{name.capitalize()}:\n {format_info(category)}"
                for name, category in self._info_categories().items()
                if category
            ]
        )

    def _get_parameter(self, name):
        v = self._parameters.get(name)
        if isinstance(v, dict):
            return RedisDictWrapper(name, self._parameters.get, self._set_parameter)
        return v

    def _set_parameter(self, name, value):
        if isinstance(value, RemoteDictWrapper):
            value = value.to_dict()
        self._parameters[name] = value

    def _del_parameter(self, name):
        self._parameters[name] = None

    def _raise_when_missing(self, *names):
        for name in names:
            if self._get_parameter(name) is None:
                raise AttributeError(f"parameter '{name}' is not set")


def add_parameter_property(cls, name):
    if hasattr(cls, name):
        return
    method = autocomplete_property(lambda self: self._get_parameter(name))
    setattr(cls, name, method)
    method = getattr(cls, name).setter(
        lambda self, value: self._set_parameter(name, value)
    )
    setattr(cls, name, method)


class RemoteDictWrapper(MutableMapping):
    """Whenever you get, set or delete the value, the entire dictionary is pushed/pull from a remote source"""

    def __dir__(self) -> Iterable[str]:
        return list(super().__dir__()) + list(self)

    def _get_all(self) -> dict:
        raise NotImplementedError

    def _set_all(self, value: Mapping) -> dict:
        raise NotImplementedError

    def to_dict(self) -> dict:
        return self._get_all()

    def __str__(self):
        return str(self._get_all())

    def __repr__(self):
        return repr(self._get_all())

    def __getitem__(self, key: str) -> Any:
        value = self._get_all()[key]
        if isinstance(value, dict):
            value = MemoryDictWrapper(self, key)
        return value

    def __setitem__(self, key: str, value: Any) -> None:
        adict = self._get_all()
        if isinstance(value, RemoteDictWrapper):
            value = value.to_dict()
        adict[key] = value
        return self._set_all(adict)

    def __delitem__(self, key: str) -> None:
        adict = self._get_all()
        del adict[key]
        return self._set_all(adict)

    def __iter__(self) -> Iterator[Any]:
        return self._get_all().__iter__()

    def __len__(self) -> int:
        return self._get_all().__len__()


class RedisDictWrapper(RemoteDictWrapper):
    def __init__(self, name: str, getter: Callable, setter: Callable) -> None:
        self._name = name
        self._getter = getter
        self._setter = setter

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'"
            )

    def __setattr__(self, name: str, value: Any) -> None:
        if name in ("_name", "_getter", "_setter"):
            return super().__setattr__(name, value)
        self[name] = value

    def _get_all(self) -> dict:
        adict = self._getter(self._name)
        if adict is None:
            return dict()
        return adict

    def _set_all(self, value: Mapping) -> None:
        self._setter(self._name, value)


class MemoryDictWrapper(RemoteDictWrapper):
    def __init__(self, parent: RemoteDictWrapper, name: str):
        self._parent = parent
        self._name = name

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'"
            )

    def __setattr__(self, name: str, value: Any) -> None:
        if name in ("_name", "_parent"):
            return super().__setattr__(name, value)
        self[name] = value

    def _get_all(self) -> dict:
        return self._parent._get_all()[self._name]

    def _set_all(self, value: Mapping) -> None:
        self._parent[self._name] = value



import numpy


def SiO2trans(E, att_position):
    """
    :param E: energy in keV
    :param att_position: attenuator position (raw bits field value)
    :returns: oyield
    """
    coeff = [-0.0030624, 0.16457, -2.708, 20.659, -80.612, 153.42, -109.27]
    ln_E = numpy.log(numpy.asarray(E))
    muL = 0.0
    for c in coeff:
        muL = muL * ln_E + c
    muL = numpy.exp(muL)  # 1/cm
    thickness = numpy.asarray(att_position) * 1.25  # cm
    return numpy.exp(-thickness * muL)
